/* 
** Copyright 2001, Michael Pfeiffer 
*/

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <dirent.h>
#include <image.h>
#include <FindDirectory.h>
#include <driver_settings.h>
#include <stdio.h>
#include <string.h>
/* this is for the standardized portion of the driver API */
/* currently only one operation is defined: B_GET_ACCELERANT_SIGNATURE */
#include <graphic_driver.h>

#include "../include/DriverInterface.h"

/* Tell the kernel what revision of the driver API we support */
int32	api_version = 2;

#define AREA_SIZE (B_PAGE_SIZE*((sizeof(struct SharedData)+B_PAGE_SIZE-1)/B_PAGE_SIZE))

#ifdef DEBUG
#define PRINT(str) dprintf(str); 
#else
#define PRINT(str) {} 
#endif

/* function prototypes defined in _KERNEL_ */
image_id load_kernel_addon(const char* name);
void     unload_kernel_addon(image_id id);

/* -----
	function pointers for the device hooks entry points
----- */
static status_t my_device_control (void* cookie, uint32 op, void* arg, size_t len);

// Implementation limitation: Only the first device (find_device()) is wrapped.

// function pointers to wrapped device driver
static status_t      (*d_init_hardware) (void) = NULL;
static status_t      (*d_init_driver) (void) = NULL;
static void          (*d_uninit_driver) (void) = NULL;
static const char**  (*d_publish_devices)() = NULL;
static device_hooks* (*d_find_device)(const char* name) = NULL;

static device_control_hook d_device_control = NULL; // of first device!!!

static char device_name[B_PATH_NAME_LENGTH+1] = ""; // of first device!!!

static device_hooks my_device_hooks = {
	NULL, 				/* -> open entry point */
	NULL, 				/* -> close entry point */
	NULL,				/* -> free cookie */
	my_device_control, 	/* -> control entry point */
	NULL,				/* -> read entry point */
	NULL,				/* -> write entry point */
	NULL,
	NULL,
	NULL,
	NULL
};


// area, so the wrapper accelerant can retrieve the name of the wrapped accelerant 
static struct SharedData *shared = NULL;
static area_id            shared_id = -1;
static image_id           did = -1;

bool init_image_symbol(image_id id, const char* name, void** func) {
    if (B_OK != get_image_symbol(id, name, B_SYMBOL_TYPE_TEXT, func)) *func = NULL;
    return *func != NULL;
}

bool find_driver(const char* root, image_id* did) {
	char driver[B_PATH_NAME_LENGTH+1];
	char path[B_PATH_NAME_LENGTH+1];
	char buffer[B_PATH_NAME_LENGTH+1];
	DIR* dir;
	image_id id;
	int i = 0;
	const char** devs = NULL;	
	
	strcpy(path, root);
	strcat(path, "/kernel/drivers/dev/graphics/");
	strcat(strcat(strcpy(buffer, "wrapper: path = "), path), "\n"); PRINT(buffer);
	
	dir = opendir(path);
	if (dir) {
		struct dirent* entry;
		while ((entry = readdir(dir)) != NULL) {
			bool ok = true;
			strcat(strcat(strcpy(buffer, "wrapper: "), entry->d_name), "\n"); PRINT(buffer);
			// skip wrapper
			if (strcmp(entry->d_name, ".") == 0 || strcmp(entry->d_name, "..") == 0 || 
			    strcmp(entry->d_name, "0wrapper") == 0) continue;
			
			strcpy(driver, path);
			strcat(driver, entry->d_name);	
			// load add on
			id = load_kernel_addon(driver);
			if (id <= 0) { PRINT("wrapper: ERROR LOADING ADD_ON\n"); break; }
			PRINT("wrapper: DRIVER LOADED\n");
			ok = ok && init_image_symbol(id, "init_hardware",   (void**)&d_init_hardware);
			if (d_init_hardware == NULL) PRINT("wrapper: init_hardware not found\n");
			ok = ok && init_image_symbol(id, "init_driver",     (void**)&d_init_driver);
			if (d_init_driver == NULL) PRINT("wrapper: init_driver not found\n");
			ok = ok && init_image_symbol(id, "uninit_driver",   (void**)&d_uninit_driver);
			if (d_uninit_driver == NULL) PRINT("wrapper: uninit_driver not found\n");
			ok = ok && init_image_symbol(id, "publish_devices", (void**)&d_publish_devices);
			if (d_publish_devices == NULL) PRINT("wrapper: publish_devices not found\n");
			ok = ok && init_image_symbol(id, "find_device",     (void**)&d_find_device);
			if (ok && B_OK == d_init_hardware()) {
				PRINT("wrapper: init_hardware: ok\n");
				if (B_OK == d_init_driver()) {
					devs = d_publish_devices();
					PRINT("wrapper: init_driver ok\n");
					if (devs == NULL) {
						PRINT("wrapper: 0 devices published\n");
					} else {
						for (i = 0; devs[i] != NULL; i ++) {
							strcpy(buffer, "wrapper: ");
							strcat(buffer, devs[i]);
							strcat(buffer, "\n");
							PRINT(buffer);
						}
						if (i > 0) {
							closedir(dir);
							return true;
						}
					}
					d_uninit_driver();
				}
			} else {
				d_init_hardware   = NULL;
				d_init_driver     = NULL;
				d_uninit_driver   = NULL;
				d_publish_devices = NULL;
				d_find_device     = NULL;
				if (ok) {
					PRINT("wrapper: call to init_hardware returned with error\n");
				}
			}
			unload_kernel_addon(id);	
		}		
		closedir(dir);
	}
	return false;
}

static status_t get_driver_param(const char* label) {
	bool p;
	/* ~/config/settings/kernel/drivers/gd_wrapper */
	void* s = load_driver_settings("gd_wrapper");
	p = get_driver_boolean_parameter(s, label, false, true);
	unload_driver_settings(s);
	return p;
}

static status_t is_driver_enabled() {
	PRINT("wrapper: is_driver_enabled\n");
	return get_driver_param("disabledriver")  ? B_ERROR : B_OK;
}


static status_t is_enabled() {
	PRINT("wrapper: enter is_enabled\n");
	return get_driver_param("changerefreshrate")  ? B_OK : B_ERROR;
}


/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
status_t
init_hardware (void)
{
	return is_driver_enabled();
}

/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
status_t
init_driver (void)
{
	/* create shared area where the name of the wrapped accelerant is stored */
	shared_id = create_area(AREA_NAME, (void**)&shared, B_ANY_ADDRESS, AREA_SIZE, B_NO_LOCK, B_READ_AREA | B_WRITE_AREA);
	if (shared_id == B_BAD_VALUE || shared_id == B_NO_MEMORY || shared_id == B_ERROR) {
		PRINT("wrapper: create_area error\n");			
		return B_ERROR;
	}
	if (find_driver("/boot/home/config/add-ons", &did) || 
		find_driver("/boot/beos/system/add-ons", &did)) {
		PRINT("wrapper: d_init_driver ok\n");			
		return B_OK;
	} else {
		delete_area(shared_id); shared_id = -1; shared = NULL;
		return B_ERROR;
	}
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
void
uninit_driver (void)
{
	PRINT("wrapper: enter uninit_driver\n");
	if (d_uninit_driver) d_uninit_driver();
	if (shared != NULL) {
		delete_area(shared_id);
		shared = NULL;
		shared_id = -1;
	}
	if (did != -1) {
		unload_kernel_addon(did); did = -1;
	}
}

	
/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
	PRINT("wrapper: enter publish_devices\n");
	if (d_publish_devices) return d_publish_devices();
	return NULL;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks*
find_device(const char* name)
{
	device_hooks* hooks = d_find_device(name);
	PRINT("wrapper: enter find_devices\n");

	if (hooks && *device_name == 0) { 
		/* wrap the first device (and only device_control hook) */
		strcpy(device_name, name);
		d_device_control = hooks->control;
		my_device_hooks.open     = hooks->open;
		my_device_hooks.read     = hooks->read;
		my_device_hooks.write    = hooks->write;
		my_device_hooks.close    = hooks->close;
		my_device_hooks.free     = hooks->free;
		my_device_hooks.select   = hooks->select;
		my_device_hooks.deselect = hooks->deselect;
		my_device_hooks.readv    = hooks->readv;
		my_device_hooks.writev   = hooks->writev;
		return &my_device_hooks;	
	} else {
		if (*device_name != 0) {
			PRINT("wrapper: can not handle more than one device\n");
		}
	}

	return hooks;
}

int atoi(const char*);

static const char *configModes[] = {
	"refreshrate640x480",
	"refreshrate800x600",
	"refreshrate1024x768",
	"refreshrate1152x900",
	"refreshrate1280x1024",
	"refreshrate1600x1200",
	NULL
};

static status_t
get_refresh_rate(void* arg, size_t len) {
	PRINT("wrapper: enter get_refresh_rate\n");
	if (arg != NULL) {
		int *res = (int*) arg;
		const char *param;
		int mode = *res;
		if (mode >= 0 && mode < 6) {
			void* s = load_driver_settings("gd_wrapper");
			param = get_driver_parameter(s, configModes[mode], "60", "60");
			unload_driver_settings(s);
			PRINT("wrapper: get_refresh_rate\n");
			PRINT(param);
			*res = atoi(param);		
			return B_OK;
		}
	} 
	return B_ERROR;
}

/* ----------
	my_device_control - handle ioctl calls
----- */

static status_t
my_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	status_t err;
	char *sig = (char *)arg;

	switch(op) {
	case B_GET_ACCELERANT_SIGNATURE:
		/* The wrapper accelerant has to load this accelerant and delegate function calls to it. */
		PRINT("wrapper: device_control GET_ACCELERANT_SIGNATURE\n");
		strcpy(sig, ACCELERANT_NAME);		
		PRINT(sig);
		/* get accelerant name and store it in area */
		err = d_device_control(cookie, op, shared->accelerant, AREA_SIZE);
		if (err != B_OK) PRINT("wrapper: d_device_control return error\n");
		PRINT(shared->accelerant);
		/* err = d_device_control(cookie, op, arg, len); */
		return err;
	case GDW_DETECT_DRIVER: 
		return B_OK;
	case GDW_FULLSCREEN_REFRESHRATE: 
		return get_refresh_rate(arg, len);
	case GDW_IS_ENABLED: 
		return is_enabled();
	default: 
		return d_device_control(cookie, op, arg, len);
	}
}

